package io.data2viz.scale

import kotlin.math.min

// TODO move to array module
/**
 * Returns the insertion point for x in array to maintain sorted order.
 * The arguments lo and hi may be used to specify a subset of the array which should be considered;
 * by default the entire array is used. If x is already present in array, the insertion point will be
 * after (to the right of) any existing entries of x in array.
 * The returned insertion point i partitions the array into two halves so that all v <= x for v in array.slice(lo, i)
 * for the left side and all v > x for v in array.slice(i, hi) for the right side.
 */
fun <T> bisect(list: List<T>, x: T, comparator: Comparator<T>, low: Int = 0, high: Int = list.size): Int {
    var lo = low
    var hi = high
    while (lo < hi) {
        val mid = (lo + hi) / 2
        if (comparator.compare(list[mid], x) > 0)
            hi = mid
        else
            lo = mid + 1
    }
    return lo
}

fun <T> bisectLeft(list: List<T>, x: T, comparator: Comparator<T>, low: Int = 0, high: Int = list.size): Int {
    var lo = low
    var hi = high
    while (lo < hi) {
        val mid = (lo + hi) / 2
        if (comparator.compare(list[mid], x) < 0)
            lo = mid + 1
        else
            hi = mid
    }
    return lo
}

// uninterpolate  [value A .. value B] --> [0 .. 1]
// interpolate [0 .. 1] --> [value A .. value B]

// TODO RGB continuous scale, HCL ...
/**
 * Continuous scales map a continuous, quantitative input domain to a continuous output range.
 * If the range is also numeric, the mapping may be inverted.
 * A continuous scale is not constructed directly; instead, try a linear, power, log,
 * identity, time or sequential color scale.
 */
abstract class ContinuousScale<R>(
        val interpolateRange: (R, R) -> (Double) -> R,
        val uninterpolateRange: ((R, R) -> (R) -> Double)? = null,
        val rangeComparator: Comparator<R>? = null) :
        RangeableScale<Double, R>,
        ClampableScale<Double, R>,
        InvertableScale<Double, R>,
        TickableScale<Double, R> {

    private var input: ((R) -> Double)? = null
    private var output: ((Double) -> R)? = null

    protected val _domain: MutableList<Double> = arrayListOf(.0, 1.0)
    protected val _range: MutableList<R> = arrayListOf()

    abstract fun interpolateDomain(from: Double, to: Double): (Double) -> Double
    abstract fun uninterpolateDomain(from: Double, to: Double): (Double) -> Double

    override operator fun invoke(domainValue: Double): R {
        if (output == null) {
            check(_domain.size == _range.size, { "Domains (in) and Ranges (out) must have the same size." })
            val uninterpolateFunc = if (clamp) uninterpolateClamp(::uninterpolateDomain) else ::uninterpolateDomain
            output =
                    if (_domain.size > 2 || _range.size > 2) polymap(uninterpolateFunc, interpolateRange)
                    else bimap(uninterpolateFunc, interpolateRange)
        }

        return output?.invoke(domainValue) ?: throw IllegalStateException()
    }

    // TODO : wrong : clamping is done on interpolateRange function...
    override fun invert(rangeValue: R): Double {
        checkNotNull(uninterpolateRange, { "No de-interpolation function for range has been found for this scale. Invert operation is impossible." })

        if (input == null) {
            check(_domain.size == _range.size, { "Domains (in) and Ranges (out) must have the same size." })
            val interpolateFunc = if (clamp) interpolateClamp(::interpolateDomain) else ::interpolateDomain
            input =
                    if (_domain.size > 2 || _range.size > 2) polymapInvert(interpolateFunc, uninterpolateRange!!)
                    else bimapInvert(interpolateFunc, uninterpolateRange!!)
        }

        return input?.invoke(rangeValue) ?: throw IllegalStateException()
    }

    override fun ticks(count: Int): List<Double> {
        return io.data2viz.core.ticks(_domain.first(), _domain.last(), count) as List<Double>
    }

    protected fun rescale() {
        input = null
        output = null
    }

    private fun uninterpolateClamp(uninterpolateFunction: (Double, Double) -> (Double) -> Double): (Double, Double) -> (Double) -> Double {
        return fun(a: Double, b: Double): (Double) -> Double {
            val d = uninterpolateFunction(a, b)
            return fun(value: Double): Double {
                return when {
                    (value <= a) -> .0
                    (value >= b) -> 1.0
                    else -> d(value)
                }
            }
        }
    }

    private fun interpolateClamp(interpolateFunction: (Double, Double) -> (Double) -> Double): (Double, Double) -> (Double) -> Double {
        return fun(a: Double, b: Double): (Double) -> Double {
            val r = interpolateFunction(a, b)
            return fun(value: Double): Double = when {
                (value <= 0.0) -> a
                (value >= 1.0) -> b
                else -> r(value)
            }
        }
    }

    private fun bimap(deinterpolateDomain: (Double, Double) -> (Double) -> Double,
                      reinterpolateRange: (R, R) -> (Double) -> R): (Double) -> R {

        val d0 = _domain[0]
        val d1 = _domain[1]
        val r0 = _range[0]
        val r1 = _range[1]

        val r: (Double) -> R
        val d: (Double) -> Double

        if (d1 < d0) {
            d = deinterpolateDomain(d1, d0)
            r = reinterpolateRange(r1, r0)
        } else {
            d = deinterpolateDomain(d0, d1)
            r = reinterpolateRange(r0, r1)
        }

        return { x: Double -> r(d(x)) }
    }

    private fun bimapInvert(reinterpolateDomain: (Double, Double) -> (Double) -> Double,
                            deinterpolateRange: (R, R) -> (R) -> Double): (R) -> Double {

        checkNotNull(rangeComparator, { "No RangeComparator has been found for this scale. Invert operation is impossible." })

        val d0 = _domain[0]
        val d1 = _domain[1]
        val r0 = _range[0]
        val r1 = _range[1]

        val r: (R) -> Double
        val d: (Double) -> Double

        if (rangeComparator!!.compare(r1, r0) < 0) {
            d = reinterpolateDomain(d1, d0)
            r = deinterpolateRange(r1, r0)
        } else {
            d = reinterpolateDomain(d0, d1)
            r = deinterpolateRange(r0, r1)
        }

        return { x: R -> d(r(x)) }
    }

    private fun polymap(uninterpolateDomain: (Double, Double) -> (Double) -> Double,
                        interpolateRange: (R, R) -> (Double) -> R): (Double) -> R {

        val d0 = _domain.first()
        val d1 = _domain.last()
        val domainReversed = d1 < d0
        val domainValues = if (domainReversed) _domain.reversed() else _domain
        val rangeValues = if (domainReversed) _range.reversed() else _range

        val size = min(_domain.size, _range.size) - 1
        val domainInterpolators = Array(size, { uninterpolateDomain(domainValues[it], domainValues[it + 1]) })
        val rangeInterpolators = Array(size, { interpolateRange(rangeValues[it], rangeValues[it + 1]) })

        return { x ->
            val index = bisect<Double>(_domain, x, naturalOrder<Double>(), 1, size) - 1
            rangeInterpolators[index](domainInterpolators[index](x))
        }
    }

    private fun polymapInvert(interpolateDomain: (Double, Double) -> (Double) -> Double,
                              uninterpolateRange: (R, R) -> (R) -> Double): (R) -> Double {

        // TODO <R> instanceOf Comparable ??
        checkNotNull(rangeComparator, { "No RangeComparator has been found for this scale. Invert operation is impossible." })

        val r0 = _range.first()
        val r1 = _range.last()
        val rangeReversed = rangeComparator!!.compare(r1, r0) < 0
        val domainValues = if (rangeReversed) _domain.reversed() else _domain
        val rangeValues = if (rangeReversed) _range.reversed() else _range

        val size = min(_domain.size, _range.size) - 1
        val domainInterpolators = Array(size, { interpolateDomain(domainValues[it], domainValues[it + 1]) })
        val rangeInterpolators = Array(size, { uninterpolateRange(rangeValues[it], rangeValues[it + 1]) })

        return { y ->
            val index = bisect<R>(rangeValues, y, rangeComparator, 1, size) - 1
            domainInterpolators[index](rangeInterpolators[index](y))
        }
    }
}